# SPDX-License-Identifier: GPL-3.0-or-later
import logging
from typing import Any, Dict, Iterator, List, Optional, Type, Union

from starmap_client.models import (
    Destination,
    Mapping,
    PaginatedRawData,
    Policy,
    QueryResponse,
    QueryResponseContainer,
)
from starmap_client.providers import StarmapProvider
from starmap_client.session import StarmapBaseSession, StarmapSession

log = logging.getLogger(__name__)


Q = Union[QueryResponse, QueryResponseContainer]

API_QUERY_RESPONSE: Dict[str, Type[Q]] = {
    "v1": QueryResponse,
    "v2": QueryResponseContainer,
    "default": QueryResponseContainer,
}


class StarmapClient(object):
    """Implement the StArMap client."""

    POLICIES_PER_PAGE = 100
    """Number of policies to retrieve per call."""

    def __init__(
        self,
        url: Optional[str] = None,
        api_version: str = "v2",
        session: Optional[StarmapBaseSession] = None,
        session_params: Optional[Dict[str, Any]] = None,
        provider: Optional[StarmapProvider] = None,
    ):
        """
        Create a new StArMapClient.

        Args:
            url (str, optional)
                URL of the StArMap endpoint. Required when session is not set.

            api_version (str, optional)
                The StArMap API version. Defaults to `v2`.
            session (StarmapBaseSession, optional)
                Defines the session object to use. Defaults to `StarmapSession` when not set
            session_params (dict, optional)
                Additional keyword arguments for StarmapSession
            provider (StarmapProvider, optional):
                Object responsible to provide mappings locally. When set the client will be query it
                first and if no mapping is found the subsequent request will be made to the server.
        """
        if url is None and session is None:
            raise ValueError(
                "Cannot initialize the client without defining either an \"url\" or \"session\"."
            )
        if provider and provider.api != api_version:
            raise ValueError(
                f"API mismatch: Provider has API {provider.api} but the client expects: {api_version}"  # noqa: E501
            )
        session_params = session_params or {}
        url = url or ""  # just to make mypy happy. The URL is mandatory if session is not defined
        self.session = session or StarmapSession(url, api_version, **session_params)
        self.api_version = api_version
        self._provider = provider
        self._policies: List[Policy] = []

    def _query(self, params: Dict[str, Any]) -> Optional[Q]:
        qr = None
        if self._provider:
            qr = self._provider.query(params)
        rsp = qr or self.session.get("/query", params=params)
        if isinstance(rsp, QueryResponse) or isinstance(rsp, QueryResponseContainer):
            log.debug(
                "Returning response from the local provider %s", self._provider.__class__.__name__
            )
            return rsp
        if rsp.status_code == 404:
            log.error(f"Marketplace mappings not defined for {params}")
            return None
        rsp.raise_for_status()
        converter = API_QUERY_RESPONSE.get(self.api_version, API_QUERY_RESPONSE["default"])
        return converter.from_json(json=rsp.json())

    def query_image(self, nvr: str, **kwargs) -> Optional[Q]:
        """
        Query StArMap using an image NVR.

        Args:
            nvr (str): The image archive name or NVR.
            workflow(Workflow, optional): The desired workflow to retrieve the mappings (APIv1 Only)

        Returns:
            Q: The query result when found or None.
        """
        return self._query(params={"image": nvr, **kwargs})

    def query_image_by_name(
        self,
        name: str,
        version: Optional[str] = None,
        **kwargs,
    ) -> Optional[Q]:
        """
        Query StArMap using an image NVR.

        Args:
            name (str): The image name from NVR.
            version (str, optional): The version from NVR.
            workflow(Workflow, optional): The desired workflow to retrieve the mappings (APIv1 Only)

        Returns:
            Q: The query result when found or None.
        """
        params = {"name": name, **kwargs}
        if version:
            params.update({"version": version})
        return self._query(params=params)

    @property
    def policies(self) -> Iterator[Policy]:
        """Iterate over all Policies registered in StArMap."""
        has_next_page = True
        page = 1

        # Iterate over pagination until there is no longer a "next" URL
        while has_next_page:
            params = {"page": page, "per_page": self.POLICIES_PER_PAGE}
            res = self.session.get("policy", params=params)
            if res.status_code == 404:
                log.error("No policies registered in StArMap.")
                return
            res.raise_for_status()

            data: PaginatedRawData = res.json()
            nav = data["nav"]

            # Yield all Policy elements from the current page list
            for item in data.get("items", []):
                yield Policy.from_json(item)

            # next iteration
            has_next_page = nav.get("next") is not None
            page += 1

    def list_policies(self) -> List[Policy]:
        """
        List all Policies present in StArMap.

        Returns:
            list(Policy): List with all policies present in StArMap.
        """
        if not self._policies:
            self._policies = [p for p in self.policies]
        return self._policies

    def get_policy(self, policy_id: str) -> Optional[Policy]:
        """
        Retrieve a single policy by its ID.

        Args:
            policy_id (str): The Policy ID to retrieve from StArMap.

        Returns:
            Policy: The requested Policy when found.
        """
        rsp = self.session.get(f"/policy/{policy_id}")
        if rsp.status_code == 404:
            log.error(f"Policy not found with ID = \"{policy_id}\"")
            return None
        rsp.raise_for_status()
        return Policy.from_json(json=rsp.json())

    def list_mappings(self, policy_id: str) -> List[Mapping]:
        """
        List all mappings for a given Policy ID.

        Args:
            policy_id (str)
                Policy ID to list the mappings.

        Returns:
            List with the Mappings for the requested Policy.
        """
        res = self.get_policy(policy_id)
        if res:
            return res.mappings
        return []

    def get_mapping(self, mapping_id: str) -> Optional[Mapping]:
        """
        Retrieve a single Marketplace Mapping by its ID.

        Args:
            mapping_id (str)
                The Markeplace Mapping ID to retrieve from StArmAp.

        Returns:
            The requested Marketplace Mapping when found.
        """
        rsp = self.session.get(f"/mapping/{mapping_id}")
        if rsp.status_code == 404:
            log.error(f"Marketplace Mapping not found with ID = \"{mapping_id}\"")
            return None
        rsp.raise_for_status()
        return Mapping.from_json(json=rsp.json())

    def list_destinations(self, mapping_id: str) -> List[Destination]:
        """
        List all destinations for a given Marketplace Mapping ID.

        Args:
            mapping_id (str)
                Marketplace Mapping ID to list the mappings.

        Returns:
            List with the Destinations for the requested Mapping.
        """
        res = self.get_mapping(mapping_id)
        if res:
            return res.destinations
        return []

    def get_destination(self, destination_id: str) -> Optional[Destination]:
        """
        Retrieve a single Destination by its ID.

        Args:
            destination_id (str)
                The Destination ID to retrieve from StArmAp.

        Returns:
            The requested Destination when found.
        """
        rsp = self.session.get(f"/destination/{destination_id}")
        if rsp.status_code == 404:
            log.error(f"Destination not found with ID = \"{destination_id}\"")
            return None
        rsp.raise_for_status()
        return Destination.from_json(json=rsp.json())



# SPDX-License-Identifier: GPL-3.0-or-later
from starmap_client.providers.base import StarmapProvider  # noqa: F401
from starmap_client.providers.memory import (  # noqa: F401
    InMemoryMapProviderV1,
    InMemoryMapProviderV2,
)



from typing import Any, Dict, List, Optional

from starmap_client.models import QueryResponse, QueryResponseContainer, QueryResponseEntity
from starmap_client.providers.base import StarmapProvider
from starmap_client.providers.utils import get_image_name


class InMemoryMapProviderV1(StarmapProvider):
    """Provide in memory (RAM) QueryResponse mapping objects for APIv1."""

    api = "v1"

    def __init__(
        self, map_responses: Optional[List[QueryResponse]] = None, *args, **kwargs
    ) -> None:
        """Crete a new InMemoryMapProvider object.

        Args:
            map_responses (list, optional)
                List of QueryResponse objects to load into memory. They will be
                used by query to fetch the correct response based on name
                and workflow.
        """
        self._separator = str(kwargs.pop("separator", "+"))
        self._content: Dict[str, QueryResponse] = {}
        super(StarmapProvider, self).__init__()
        self._boostrap(map_responses)

    def _boostrap(self, map_responses: Optional[List[QueryResponse]]) -> None:
        """Initialize the internal content dictionary.

        Args:
            map_responses (list, optional)
                List of QueryResponse objects to load into memory.
        """
        if not map_responses:
            return None

        # The in memory content is made of a combination of name and workflow
        for map in map_responses:
            key = f"{map.name}{self._separator}{map.workflow.value}"
            self._content[key] = map

    def list_content(self) -> List[QueryResponse]:
        """Return a list of stored content."""
        return list(self._content.values())

    def store(self, response: QueryResponse) -> None:
        """Store/replace a single QueryResponse object.

        Args:
            response (QueryResponse):
                The object to store.
        """
        key = f"{response.name}{self._separator}{response.workflow.value}"
        self._content[key] = response

    def query(self, params: Dict[str, Any]) -> Optional[QueryResponse]:
        """Return the mapping from memory according to the received params.

        Args:
            params (dict):
                The request params to retrieve the mapping.
        Returns:
            The requested mapping when found.
        """
        name = params.get("name") or get_image_name(params.get("image"))
        workflow = str(params.get("workflow", ""))
        search_key = f"{name}{self._separator}{workflow}"
        return self._content.get(search_key)


class InMemoryMapProviderV2(StarmapProvider):
    """Provide in memory (RAM) QueryResponseContainer objects for APIv2."""

    api = "v2"

    def __init__(self, container: QueryResponseContainer, *args, **kwargs) -> None:
        """Crete a new InMemoryMapProvider object.

        Args:
            container (QueryResponseContainer)
                QueryResponseContainer object to load into memory. It will be
                used by query to return the correct response based on name
                and workflow.
        """
        self._container = container
        super(StarmapProvider, self).__init__()

    def query(self, params: Dict[str, Any]) -> Optional[QueryResponseContainer]:
        """Retrieve the mapping without using the server.

        It relies in the local provider to retrieve the correct mappings
        according to the parameters.

        Args:
            params (dict):
                The request params to retrieve the mapping.
        Returns:
            The requested container with mappings when found.
        """
        filter_params = {"name": params.get("name") or get_image_name(params.get("image"))}
        for k in ["cloud", "workflow"]:
            v = params.get(k)
            if v:
                filter_params.update({k: v})
        res = self._container.filter_by(**filter_params)
        if res:
            return QueryResponseContainer(res)
        return None

    def list_content(self) -> List[QueryResponseEntity]:
        """Return a the responses stored in the container."""
        return self._container.responses

    def store(self, response: QueryResponseEntity) -> None:
        """Store a single response into the local provider's container.

        Args:
            container (container):
                The container to store.
        """
        self._container.responses.append(response)



from abc import ABC, abstractmethod
from typing import Any, Dict, Generic, List, Optional, TypeVar

from starmap_client.models import QueryResponse, QueryResponseContainer, QueryResponseEntity

T = TypeVar("T", QueryResponse, QueryResponseContainer, QueryResponseEntity)


class StarmapProvider(ABC, Generic[T]):
    """Define the interface for a local mappings provider."""

    api = "default"
    """The provider's API level implementation."""

    @abstractmethod
    def query(self, params: Dict[str, Any]) -> Optional[T]:
        """Retrieve the mapping without using the server.

        It relies in the local provider to retrieve the correct mapping
        according to the parameters.

        Args:
            params (dict):
                The request params to retrieve the mapping.
        Returns:
            The requested mapping when found.
        """

    @abstractmethod
    def list_content(self) -> List[T]:
        """Return a list with all stored responses."""

    @abstractmethod
    def store(self, response: T) -> None:
        """Store a single response into the local provider.

        Args:
            response (response):
                The object to store.
        """



# SPDX-License-Identifier: GPL-3.0-or-later
# The functions below were adapted from Kobo's RPMLib:
# https://github.com/release-engineering/kobo/blob/master/kobo/rpmlib.py
import logging
from typing import Dict, Optional, Tuple

log = logging.getLogger(__name__)


def split_nvr_epoch(nvre: str) -> Tuple[str, str]:
    """
    Split nvre to N-V-R and E.

    :param nvre: E:N-V-R or N-V-R:E string
    :return: (N-V-R, E)
    """
    if ":" in nvre:
        log.debug("Splitting NVR and Epoch")
        if nvre.count(":") != 1:
            raise RuntimeError(f"Invalid NVRE: {nvre}")

        nvr, epoch = nvre.rsplit(":", 1)
        if "-" in epoch:
            if "-" not in nvr:
                # switch nvr with epoch
                nvr, epoch = epoch, nvr
            else:
                # it's probably N-E:V-R format, handle it after the split
                nvr, epoch = nvre, ""
    else:
        log.debug("No epoch to split")
        nvr, epoch = nvre, ""

    return nvr, epoch


def parse_nvr(nvre: str) -> Dict[str, str]:
    """
    Split N-V-R into a dictionary.

    :param nvre: N-V-R:E, E:N-V-R or N-E:V-R string
    :return: {name, version, release, epoch}
    """
    log.debug("Parsing NVR")
    if "/" in nvre:
        nvre = nvre.split("/")[-1]

    nvr, epoch = split_nvr_epoch(nvre)

    log.debug("Splitting NVR parts")
    nvr_parts = nvr.rsplit("-", 2)
    if len(nvr_parts) != 3:
        raise RuntimeError(f"Invalid NVR: {nvr}")

    # parse E:V
    if epoch == "" and ":" in nvr_parts[1]:
        log.debug("Parsing E:V")
        epoch, nvr_parts[1] = nvr_parts[1].split(":", 1)

    # check if epoch is empty or numeric
    if epoch != "":
        try:
            int(epoch)
        except ValueError:
            raise RuntimeError(f"Invalid epoch '{epoch}' in '{nvr}'")

    result = dict(zip(["name", "version", "release"], nvr_parts))
    result["epoch"] = epoch
    return result


def get_image_name(image: Optional[str]) -> str:
    """Retrieve the name from NVR."""
    if not image:
        return ""
    nvr = parse_nvr(image)
    return nvr.get("name", "")



from setuptools import setup, find_packages

setup(
    name='starmap-client',
    description='Client for StArMap',
    version='2.0.0',
    keywords='stratosphere content artifact mapping cli',
    author='Jonathan Gangi',
    author_email='jgangi@redhat.com',
    url='https://github.com/release-engineering/starmap-client',
    license='GPLv3+',
    packages=find_packages(exclude=['tests', 'tests.*']),
    include_package_data=True,
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    install_requires=[
        'attrs',
        'requests',
        'requests_mock',
        'urllib3<2.0.0'
    ],
    zip_safe=False,
)



from typing import Any, Dict


def assert_is_dict(data: Any) -> None:
    """Ensure the incoming data is a dictionary, raises ``ValueError`` if not."""
    if not isinstance(data, dict):
        raise ValueError(f"Expected dictionary, got {type(data)}: {data}")


def dict_merge(a: Dict[str, Any], b: Dict[str, Any]) -> Dict[str, Any]:
    """Return a new dictionary with the combination of A and B.

    Args:
        a (dict):
            The left dictionary to be combined
        b (dict):
            The right dictionary to be combined. It will override the same keys from A.
    Returns:
        dict: A new dictionary with combination of A and B.
    """
    for x in [a, b]:
        assert_is_dict(x)
    return a | b



# SPDX-License-Identifier: GPL-3.0-or-later
import logging
import re
from abc import ABC, abstractmethod
from typing import Any, Dict

import requests
import requests_mock
from requests.adapters import HTTPAdapter, Retry

log = logging.getLogger(__name__)


class StarmapBaseSession(ABC):
    """Define the interface for the Starmap's session objects."""

    @abstractmethod
    def get(self, path: str, **kwargs) -> requests.Response:
        """Perform a GET request on StArMap."""

    @abstractmethod
    def post(self, path: str, json: Dict[str, Any], **kwargs) -> requests.Response:
        """Perform a POST request on StArMap."""

    @abstractmethod
    def put(self, path: str, json: Dict[str, Any], **kwargs) -> requests.Response:
        """Perform a PUT request on StArMap."""


class StarmapSession(StarmapBaseSession):
    """Implement a HTTP(S) session with StArMap."""

    def __init__(
        self,
        url: str,
        api_version: str,
        retries: int = 3,
        backoff_factor: float = 2.0,
    ):
        """
        Create the StarmapSession object.

        Args:
            url (str)
                The StArMap server endpoint base URL
            api_version
                The StArMap server API version to call
            retries (int, optional)
                The number of request retries on failure
            backoff_factor (float, optional)
                The backoff factor to apply between attempts after the second try
        """
        super(StarmapSession, self).__init__()
        self.url = url
        self.api_version = api_version
        self.session = requests.Session()
        retry = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=set(range(500, 512)),
        )
        adapter = HTTPAdapter(max_retries=retry)
        self.session.mount("https://", adapter)
        self.verify = True

    def _request(self, method: str, path: str, **kwargs) -> requests.Response:
        """Perform a generic request on StArMap."""
        headers = {
            "Accept": "application/json",
        }

        log.info(f"Sending a {method} request to {path}")
        url_elements = [self.url, f"/api/{self.api_version}", path]
        url = "/".join(arg.strip("/") for arg in url_elements)

        return self.session.request(method, url=url, headers=headers, verify=self.verify, **kwargs)

    def get(self, path: str, **kwargs) -> requests.Response:
        """Perform a GET request on StArMap."""
        return self._request("get", path, **kwargs)

    def post(self, path: str, json: Dict[str, Any], **kwargs) -> requests.Response:
        """Perform a POST request on StArMap."""
        return self._request("post", path, json=json, **kwargs)

    def put(self, path: str, json: Dict[str, Any], **kwargs) -> requests.Response:
        """Perform a PUT request on StArMap."""
        return self._request("put", path, json=json, **kwargs)


class StarmapMockSession(StarmapSession):
    """Implement a mock session with predefined responses."""

    def __init__(self, url: str, api_version: str, status_code: int = 404, json_data: Any = None):
        """Create the StarmapMockSession object.

        Args:
            url (str)
                The mock server endpoint base URL
            api_version
                The mock server API version to call
            status_code (optional, int)
                The status code to return on each request

            json_data (optional, any)
                The JSON data to return on each request
        """
        super(StarmapMockSession, self).__init__(url, api_version)
        self.url = f"mock://{url}"
        self.api_version = api_version
        self.status_code = status_code
        self.json_data = json_data or {}
        self.session = requests.Session()
        self.adapter = requests_mock.Adapter()
        self.session.mount("mock://", self.adapter)
        self._register_starmap_endpoints()

    def _register_starmap_endpoints(self) -> None:
        base_url_elements = [self.url, f"/api/{self.api_version}"]
        base_url = "/".join(arg.strip("/") for arg in base_url_elements)
        methods = ["GET", "POST", "PUT"]
        for m in methods:
            self.register_uri(m, re.compile(f"{base_url}/.*"))  # type: ignore [arg-type]

    def register_uri(self, method: str, uri: str):
        """Register an URI into the ``requests_mock`` adapter.

        Args:
            method (str):
                The HTTP method to register
            uri:
                The URI to register
        """
        self.adapter.register_uri(
            method, url=uri, status_code=self.status_code, json=self.json_data
        )



# SPDX-License-Identifier: GPL-3.0-or-later
import sys
from copy import deepcopy
from enum import Enum
from typing import Any, Dict, List, Optional, Type

if sys.version_info >= (3, 8):
    from typing import TypedDict  # pragma: no cover
else:
    from typing_extensions import TypedDict  # pragma: no cover

from attrs import Attribute, Factory, asdict, evolve, field, frozen
from attrs.validators import deep_iterable, deep_mapping, instance_of, min_len, optional

from starmap_client.utils import assert_is_dict, dict_merge

__all__ = [
    'BillingCodeRule',
    'BillingImageType',
    'Destination',
    'Mapping',
    'Policy',
    'QueryResponse',
    'QueryResponseEntity',
    'QueryResponseContainer',
    'PaginatedRawData',
    'PaginationMetadata',
    'Workflow',
]


# ============================================ Common ==============================================


class PaginationMetadata(TypedDict):
    """Datastructure of the metadata about the paginated query."""

    first: str
    last: str
    next: Optional[str]
    page: int
    per_page: int
    previous: Optional[str]
    total: int
    total_pages: int


class PaginatedRawData(TypedDict):
    """Represent a paginated StArMap data in its raw format (Dict)."""

    items: List[Any]
    nav: PaginationMetadata


class Workflow(str, Enum):
    """Define the valid workflows for StArMap."""

    community = "community"

    stratosphere = "stratosphere"


@frozen
class StarmapJSONDecodeMixin:
    """Implement the default JSON deserialization for StArMap models."""

    @classmethod
    def _assert_json_dict(cls, json: Any) -> None:
        """
        Ensure the given JSON is an instance of `dict`.

        Args:
            json (Any)
                A JSON containing a StArMap response.
        """
        if not isinstance(json, dict):
            raise ValueError(
                f"Got an unsupported JSON type: \"{type(json)}\". Expected: \"<class 'dict'>\'"
            )

    @classmethod
    def _preprocess_json(cls, json: Dict[str, Any]) -> Dict[str, Any]:
        """
        Preprocess the JSON before converting it to class object.

        It's intended to be overriden by base classes which needs to do it.

        Args:
            json (dict)
                A JSON containing a StArMap response.
        Returns:
            dict: The modified JSON.
        """
        return json

    @classmethod
    def from_json(cls, json: Any):
        """
        Convert a JSON dictionary into class object.

        Args:
            json (dict)
                A JSON containing a StArMap response.
        Returns:
            The converted object from JSON.
        """
        cls._assert_json_dict(json)

        json = cls._preprocess_json(json)

        args = {}
        cls_attr = [a.name for a in cls.__attrs_attrs__ if isinstance(a, Attribute)]  # type: ignore
        for a in cls_attr:
            args[a] = json.pop(a, None)
        return cls(**args)


@frozen
class MetaMixin:
    """Mixin for defining the meta attribute and its validator."""

    meta: Optional[Dict[str, Any]] = field()
    """Dictionary with additional information related to a VM image."""

    @meta.validator
    def _is_meta_dict_of_str_any(self, attribute: Attribute, value: Any):
        if not value:
            return None
        if not isinstance(value, dict):
            raise TypeError(f"The value for \"{attribute.name}\" should be a dictionary.")
        for k in value.keys():
            if not isinstance(k, str):
                raise ValueError(f"Invalid key \"{k}\" for \"{attribute.name}\". Expected: \"str\"")


@frozen
class StarmapBaseData(MetaMixin, StarmapJSONDecodeMixin):
    """Represent the common data present in StArMap entities."""

    id: Optional[str] = field(validator=optional(instance_of(str)))
    """
    The unique ID for a StArMap model.
    This field is never set on :class:`~starmap_client.models.QueryResponse`.
    """


@frozen
class Destination(StarmapBaseData):
    """Represent a destination entry from Mapping."""

    architecture: Optional[str] = field(validator=optional(instance_of(str)))
    """Architecture of the VM image."""

    destination: str = field(validator=instance_of(str))
    """The product listing destination in the cloud marketplace."""

    overwrite: bool = field(validator=instance_of(bool))
    """Whether to replace the existing VM image in the destination or append."""

    restrict_version: bool = field(validator=instance_of(bool))
    """Whether to restrict and image and delete it's AMI and snapshot"""

    restrict_major: Optional[int] = field(validator=optional(instance_of(int)))
    """How many major versions are allowed in product"""

    restrict_minor: Optional[int] = field(validator=optional(instance_of(int)))
    """How many minor versions are allowed in product"""

    provider: Optional[str] = field(validator=optional(instance_of(str)))
    """Represent the RHSM provider name for the community workflow."""

    tags: Optional[Dict[str, str]] = field(
        validator=optional(
            deep_mapping(
                key_validator=instance_of(str),
                value_validator=instance_of(str),
                mapping_validator=instance_of(dict),
            )
        )
    )
    """Dictionary with custom tags to be set on cloud marketplaces resources."""


@frozen
class Mapping(StarmapBaseData):
    """Represent a marketplace Mapping from Policy."""

    destinations: List[Destination] = field(
        validator=[
            min_len(1),
            deep_iterable(
                member_validator=instance_of(Destination), iterable_validator=instance_of(list)
            ),
        ],
        converter=lambda x: [Destination.from_json(d) for d in x] if x else [],
    )
    """List of destinations for the marketplace account."""

    marketplace_account: str = field(validator=instance_of(str))
    """A string representing the destination marketplace account."""

    version_fnmatch: Optional[str] = field(validator=optional(instance_of(str)))
    """A ``fnmatch`` string to apply the destinations only to the matched NVR versions.

    It can't be set together with ``version_regexmatch``.
    """

    version_regexmatch: Optional[str] = field(validator=optional(instance_of(str)))
    """A ``regex`` string to apply the destinations only to the matched NVR versions.

    It can't be set together with ``version_fnmatch``."""


@frozen
class Policy(StarmapBaseData):
    """Represent a StArMap policy."""

    mappings: List[Mapping] = field(
        validator=[
            min_len(1),
            deep_iterable(
                member_validator=instance_of(Mapping), iterable_validator=instance_of(list)
            ),
        ],
        converter=lambda x: [Mapping.from_json(m) for m in x] if x else [],
    )
    """List of marketplace mappings which the Policy applies to."""

    name: str = field(validator=instance_of(str))
    """The Koji Package name representing also the Policy name."""

    workflow: Workflow = field(converter=lambda x: Workflow(x))
    """The policy workflow name."""


# ============================================ APIv1 ===============================================


@frozen
class QueryResponse(StarmapJSONDecodeMixin):
    """Represent a query response from StArMap."""

    name: str = field(validator=instance_of(str))
    """The :class:`~Policy` name."""

    workflow: Workflow = field(converter=lambda x: Workflow(x))
    """The :class:`~Policy` workflow."""

    clouds: Dict[str, List[Destination]] = field(
        default=Factory(dict),
        validator=deep_mapping(
            key_validator=instance_of(str),
            value_validator=deep_iterable(
                member_validator=instance_of(Destination), iterable_validator=instance_of(list)
            ),
            mapping_validator=instance_of(dict),
        ),
    )
    """Dictionary with the cloud marketplaces aliases and their respective Destinations."""

    @classmethod
    def _preprocess_json(cls, json: Any) -> Dict[str, Any]:
        """
        Convert the JSON format to the expected by QueryResponse.

        Params:
            json (dict): A JSON containing a StArMap Query response.
        Returns:
            dict: The modified JSON.
        """
        mappings = json.pop("mappings", {})
        for c in mappings.keys():
            if not isinstance(mappings[c], list):
                raise ValueError(f"Expected mappings to be a list, got \"{type(mappings[c])}\".")
            dst = [Destination.from_json(d) for d in mappings[c]]
            mappings[c] = dst
        json["clouds"] = mappings
        return json


# ============================================ APIv2 ===============================================


class BillingImageType(str, Enum):
    """Define the image type for :class:`~BillingCodeRule` for APIv2."""

    access = "access"
    hourly = "hourly"
    marketplace = "marketplace"


@frozen
class BillingCodeRule(StarmapJSONDecodeMixin):
    """Define a single Billing Code Configuration rule for APIv2."""

    codes: List[str] = field(
        validator=deep_iterable(
            member_validator=instance_of(str), iterable_validator=instance_of(list)
        )
    )
    """The billing codes to insert when this rule is matched."""

    image_name: str = field(validator=instance_of(str))
    """The image name to match the rule."""

    image_types: List[BillingImageType] = field(
        converter=lambda x: [BillingImageType[d] for d in x],
        validator=deep_iterable(
            member_validator=instance_of(BillingImageType), iterable_validator=instance_of(list)
        ),
    )
    """Image types list. Supported values are ``access`` and ``hourly``."""

    name: Optional[str]
    """The billing code rule name."""


@frozen
class MappingResponseObject(StarmapJSONDecodeMixin, MetaMixin):
    """Represent a single mapping response from :class:`~QueryResponseObject` for APIv2."""

    destinations: List[Destination] = field(
        validator=deep_iterable(
            iterable_validator=instance_of(list), member_validator=instance_of(Destination)
        )
    )
    """List of destinations for the mapping response object."""

    provider: Optional[str] = field(validator=optional(instance_of(str)))
    """The provider name for the community workflow."""

    @staticmethod
    def _unify_meta_with_destinations(json: Dict[str, Any]) -> None:
        """Merge the ``meta`` data from mappings into the destinations."""
        destinations = json.get("destinations", [])
        if not isinstance(destinations, list):
            raise ValueError(f"Expected destinations to be a list, got \"{type(destinations)}\"")
        meta = json.get("meta", {})
        for d in destinations:
            d["meta"] = dict_merge(meta, d.get("meta", {}))

    @classmethod
    def _preprocess_json(cls, json: Any) -> Dict[str, Any]:
        """
        Properly adjust the Destinations list for building this object.

        Params:
            json (dict): A JSON containing a StArMap Query response.
        Returns:
            dict: The modified JSON.
        """
        cls._unify_meta_with_destinations(json)
        provider = json.get("provider", None)
        destinations = json.get("destinations", [])
        converted_destinations = []
        for d in destinations:
            d["provider"] = provider
            converted_destinations.append(Destination.from_json(d))
        json["destinations"] = converted_destinations
        return json


@frozen
class QueryResponseEntity(StarmapJSONDecodeMixin, MetaMixin):
    """Represent a single query response entity from StArMap APIv2."""

    name: str = field(validator=instance_of(str))
    """The :class:`~Policy` name."""

    billing_code_config: Optional[Dict[str, BillingCodeRule]] = field(
        validator=optional(
            deep_mapping(
                key_validator=instance_of(str),
                value_validator=instance_of(BillingCodeRule),
                mapping_validator=instance_of(dict),
            )
        )
    )
    """The Billing Code Configuration for the community workflow."""

    cloud: str = field(validator=instance_of(str))
    """The cloud name where the destinations are meant to."""

    workflow: Workflow = field(converter=lambda x: Workflow(x))
    """The :class:`~Policy` workflow."""

    mappings: Dict[str, MappingResponseObject] = field(
        validator=deep_mapping(
            key_validator=instance_of(str),
            value_validator=instance_of(MappingResponseObject),
            mapping_validator=instance_of(dict),
        ),
    )
    """Dictionary with the cloud account names and MappingResponseObjects."""

    @property
    def account_names(self) -> List[str]:
        """Return the list of cloud account names declared on ``mappings``."""
        return list(self.mappings.keys())

    @property
    def all_mappings(self) -> List[MappingResponseObject]:
        """Return all ``MappingResponseObject`` stored in ``mappings``."""
        return list(self.mappings.values())

    def get_mapping_for_account(self, account: str) -> MappingResponseObject:
        """Return a single ``MappingResponseObject`` for a given account name.

        Args:
            account (str):
                The account name to retrieve the ``MappingResponseObject``
        Returns:
            MappingResponseObject: The required mapping when found
        Raises: KeyError when not found
        """
        obj = self.mappings.get(account, None)
        if not obj:
            raise KeyError(f"No mappings found for account name {account}")
        return obj

    def to_classic_query_response(self) -> QueryResponse:
        """Return the representation of this object as a :class:`~QueryResponse` from APIv1."""

        def add_bc_to_dst_meta(clouds: Dict[str, List[Destination]]):
            if self.billing_code_config:
                bc_data = {k: asdict(v) for k, v in self.billing_code_config.items()}
                for dst_list in clouds.values():
                    for d in dst_list:
                        meta = d.meta or {}
                        meta["billing-code-config"] = bc_data
                        d = evolve(d, meta=meta)

        clouds: Dict[str, List[Destination]] = {}
        for k, v in self.mappings.items():
            clouds[k] = [deepcopy(d) for d in v.destinations]

        if self.billing_code_config:
            add_bc_to_dst_meta(clouds)

        return QueryResponse(name=self.name, workflow=self.workflow, clouds=clouds)

    @staticmethod
    def _unify_meta_with_mappings(json: Dict[str, Any]) -> None:
        """Merge the ``meta`` data from package into the mappings."""
        mappings = json.get("mappings", {})
        meta = json.get("meta", {})
        for k, v in mappings.items():
            mappings[k]["meta"] = dict_merge(meta, v.get("meta", {}))

    @classmethod
    def _preprocess_json(cls, json: Dict[str, Any]) -> Dict[str, Any]:
        """
        Properly adjust the MappingResponseObject list and BillingCodeConfig dict for building this object.

        Params:
            json (dict): A JSON containing a StArMap Query response.
        Returns:
            dict: The modified JSON.
        """  # noqa: D202 E501

        def parse_entity_build_obj(
            entity_name: str, converter_type: Type[StarmapJSONDecodeMixin]
        ) -> None:
            entity = json.pop(entity_name, {})
            for k in entity.keys():
                assert_is_dict(entity[k])
                obj = converter_type.from_json(entity[k])
                entity[k] = obj
            json[entity_name] = entity

        bcc = json.pop("billing-code-config", {})
        json["billing_code_config"] = bcc
        cls._unify_meta_with_mappings(json)
        parse_entity_build_obj("mappings", MappingResponseObject)
        parse_entity_build_obj("billing_code_config", BillingCodeRule)
        return json


@frozen
class QueryResponseContainer:
    """Represent a full query response from APIv2."""

    responses: List[QueryResponseEntity] = field(
        validator=deep_iterable(
            member_validator=instance_of(QueryResponseEntity), iterable_validator=instance_of(list)
        )
    )
    """List with all responses from a Query V2 mapping."""

    @classmethod
    def from_json(cls, json: Any):
        """
        Convert the APIv2 response JSON into this object.

        Args:
            json (list)
                A JSON containing a StArMap APIv2 response.
        Returns:
            The converted object from JSON.
        """
        if not isinstance(json, list):
            raise ValueError(f"Expected root to be a list, got \"{type(json)}\".")

        responses = [QueryResponseEntity.from_json(qre) for qre in json]
        return cls(responses)

    def filter_by_name(
        self, name: str, responses: Optional[List[QueryResponseEntity]] = None
    ) -> List[QueryResponseEntity]:
        """Return a sublist of the responses with only the selected image name.

        Args:
            name (str):
                The image name to filter the list of responses
            responses (list, optional):
                List of existing responses to filter. Default by the container's own list.
        Returns:
            list: The sublist with only the selected image name.
        """
        if responses == []:
            return responses
        rsp = responses or self.responses
        return [x for x in rsp if x.name == name]

    def filter_by_workflow(
        self, workflow: Workflow, responses: Optional[List[QueryResponseEntity]] = None
    ) -> List[QueryResponseEntity]:
        """Return a sublist of the responses with only the selected workflow.

        Args:
            workflow (Workflow):
                The workflow to filter the list of responses
            responses (list, optional):
                List of existing responses to filter. Default by the container's own list.
        Returns:
            list: The sublist with only the selected workflows
        """
        if responses == []:
            return responses
        rsp = responses or self.responses
        return [x for x in rsp if x.workflow == workflow]

    def filter_by_cloud(
        self, cloud: str, responses: Optional[List[QueryResponseEntity]] = None
    ) -> List[QueryResponseEntity]:
        """Return a sublist of the responses with only the selected cloud name.

        Args:
            cloud (str):
                The cloud name to filter the list of responses
            responses (list, optional):
                List of existing responses to filter. Default by the container's own list.
        Returns:
            list: The sublist with only the selected cloud name.
        """
        if responses == []:
            return responses
        rsp = responses or self.responses
        return [x for x in rsp if x.cloud == cloud]

    def filter_by(self, **kwargs) -> List[QueryResponseEntity]:
        """Return a sublist of the responses with the selected filters."""
        filters = {
            "name": self.filter_by_name,
            "cloud": self.filter_by_cloud,
            "workflow": self.filter_by_workflow,
        }
        res = self.responses
        for k, v in kwargs.items():
            res = filters[k](v, responses=res)
        return res
